import asyncio
import hashlib
from functools import partial
from pathlib import Path

from mootdx.financial import financial
from mootdx.logger import logger
from mootdx.utils import TqdmUpTo


def download(downdir, filename):
    """
    带进度条下载函数
    :param downdir:
    :param filename:
    :return:
    """

    with TqdmUpTo(unit='B', unit_scale=True, miniters=1, ascii=True) as t:
        financial.Financial().fetch_only(report_hook=t.update_to, filename=filename, downdir=downdir)

    return True


async def fetch_file(downdir, file_obj):
    """
    下载文件

    :param downdir:
    :param file_obj: 文件对象
    :return:
    """

    filepath = Path(downdir) / file_obj['filename']

    # 判断文件是否存在, 验证文件名和哈希值
    if filepath.exists() and file_obj['hash'] == hashlib.md5(open(filepath, 'rb').read()).hexdigest():
        logger.warning(f'文件已经存在: {filepath}')
        return None

    result = await asyncio.get_event_loop().run_in_executor(
        None,
        partial(financial.Financial().fetch_only, report_hook=None, filename=file_obj['filename'], downdir=downdir),
    )

    return result


class Affair(object):
    @staticmethod
    def parse(downdir='.', filename=None, **kwargs):
        """
        按目录解析文件

        :param downdir:
        :param filename:
        :return:
        """

        if not filename:
            logger.critical('文件名不能为空!')
            return None

        filepath = Path(downdir) / filename
        filepath.exists() or Affair.fetch(downdir, filename)

        if Path(filepath).exists():
            return financial.FinancialReader().to_data(filepath, **kwargs)

        logger.warning(f'文件不存在：{filename}')

        return None

    @staticmethod
    def files():
        """
        财务文件列表

        :return:
        """

        history = financial.FinancialList()
        results = history.fetch_and_parse()

        return results

    @staticmethod
    def fetch(downdir: str = None, filename: str = None):  # noqa
        """
        财务数据下载

        :param downdir: 下载目录
        :param filename: 文件名
        :return:
        """

        history = financial.FinancialList()
        crawler = financial.Financial()
        downdir = downdir or '.'  # noqa

        if not Path(downdir).is_dir():
            logger.warning('下载目录不存在, 进行创建.')
            Path(downdir).mkdir(parents=True)

        if filename:
            logger.debug(f'下载文件 {filename}.')

            with TqdmUpTo(unit='B', unit_scale=True, miniters=1, ascii=True) as t:
                crawler.fetch_only(report_hook=t.update_to, filename=filename, downdir=downdir)

            return True

        tasks = []
        event = asyncio.get_event_loop()

        for x in history.fetch_and_parse():
            task = event.create_task(fetch_file(file_obj=x, downdir=downdir))
            tasks.append(task)

        event.run_until_complete(asyncio.wait(tasks))



class MootdxException(Exception):
    """Base notifier exception. Catch this to catch all of :mod:`notifiers` errors"""

    def __init__(self, *args, **kwargs):
        """
        Looks for ``provider``, ``message`` and ``data`` in kwargs
        :param args: Exception arguments
        :param kwargs: Exception kwargs
        """
        self.provider = kwargs.get('provider')
        self.response = kwargs.get('response')

        self.message = kwargs.get('message')
        self.data = kwargs.get('data')

        super().__init__(self.message)

    def __repr__(self):
        return f'<MOOTDXError: {self.message}>'


class MootdxValidationException(Exception):
    def __init__(self, *args, **kwargs):
        pass


class MootdxModuleNotFoundError(Exception):
    def __init__(self, *args, **kwargs):
        pass


class FileNeedRefresh(FileNotFoundError):
    pass



columns = [
    'report_date',
    '基本每股收益',
    '扣除非经常性损益每股收益',
    '每股未分配利润',
    '每股净资产',
    '每股资本公积金',
    '净资产收益率',
    '每股经营现金流量',
    '货币资金',
    '交易性金融资产',
    '应收票据',
    '应收账款',
    '预付款项',
    '其他应收款',
    '应收关联公司款',
    '应收利息',
    '应收股利',
    '存货',
    '其中：消耗性生物资产',
    '一年内到期的非流动资产',
    '其他流动资产',
    '流动资产合计',
    '可供出售金融资产',
    '持有至到期投资',
    '长期应收款',
    '长期股权投资',
    '投资性房地产',
    '固定资产',
    '在建工程',
    '工程物资',
    '固定资产清理',
    '生产性生物资产',
    '油气资产',
    '无形资产',
    '开发支出',
    '商誉',
    '长期待摊费用',
    '递延所得税资产',
    '其他非流动资产',
    '非流动资产合计',
    '资产总计',
    '短期借款',
    '交易性金融负债',
    '应付票据',
    '应付账款',
    '预收款项',
    '应付职工薪酬',
    '应交税费',
    '应付利息',
    '应付股利',
    '其他应付款',
    '应付关联公司款',
    '一年内到期的非流动负债',
    '其他流动负债',
    '流动负债合计',
    '长期借款',
    '应付债券',
    '长期应付款',
    '专项应付款',
    '预计负债',
    '递延所得税负债',
    '其他非流动负债',
    '非流动负债合计',
    '负债合计',
    '实收资本（或股本）',
    '资本公积',
    '盈余公积',
    '减：库存股',
    '未分配利润',
    '少数股东权益',
    '外币报表折算价差',
    '非正常经营项目收益调整',
    '所有者权益（或股东权益）合计',
    '负债和所有者（或股东权益）合计',
    '其中：营业收入',
    '其中：营业成本',
    '营业税金及附加',
    '销售费用',
    '管理费用',
    '勘探费用',
    '财务费用',
    '资产减值损失',
    '加：公允价值变动净收益',
    '投资收益',
    '其中：对联营企业和合营企业的投资收益',
    '影响营业利润的其他科目',
    '三、营业利润',
    '加：补贴收入',
    '营业外收入',
    '减：营业外支出',
    '其中：非流动资产处置净损失',
    '加：影响利润总额的其他科目',
    '四、利润总额',
    '减：所得税',
    '加：影响净利润的其他科目',
    '五、净利润',
    '归属于母公司所有者的净利润',
    '少数股东损益',
    '销售商品、提供劳务收到的现金',
    '收到的税费返还',
    '收到其他与经营活动有关的现金',
    '经营活动现金流入小计',
    '购买商品、接受劳务支付的现金',
    '支付给职工以及为职工支付的现金',
    '支付的各项税费',
    '支付其他与经营活动有关的现金',
    '经营活动现金流出小计',
    '经营活动产生的现金流量净额',
    '收回投资收到的现金',
    '取得投资收益收到的现金',
    '处置固定资产、无形资产和其他长期资产收回的现金净额',
    '处置子公司及其他营业单位收到的现金净额',
    '收到其他与投资活动有关的现金',
    '投资活动现金流入小计',
    '购建固定资产、无形资产和其他长期资产支付的现金',
    '投资支付的现金',
    '取得子公司及其他营业单位支付的现金净额',
    '支付其他与投资活动有关的现金',
    '投资活动现金流出小计',
    '投资活动产生的现金流量净额',
    '吸收投资收到的现金',
    '取得借款收到的现金',
    '收到其他与筹资活动有关的现金',
    '筹资活动现金流入小计',
    '偿还债务支付的现金',
    '分配股利、利润或偿付利息支付的现金',
    '支付其他与筹资活动有关的现金',
    '筹资活动现金流出小计',
    '筹资活动产生的现金流量净额',
    '四、汇率变动对现金的影响',
    '四(2)、其他原因对现金的影响',
    '五、现金及现金等价物净增加额',
    '期初现金及现金等价物余额',
    '期末现金及现金等价物余额',
    '净利润',
    '加：资产减值准备',
    '固定资产折旧、油气资产折耗、生产性生物资产折旧',
    '无形资产摊销',
    '长期待摊费用摊销',
    '处置固定资产、无形资产和其他长期资产的损失',
    '固定资产报废损失',
    '公允价值变动损失',
    '财务费用',
    '投资损失',
    '递延所得税资产减少',
    '递延所得税负债增加',
    '存货的减少',
    '经营性应收项目的减少',
    '经营性应付项目的增加',
    '其他',
    '经营活动产生的现金流量净额2',
    '债务转为资本',
    '一年内到期的可转换公司债券',
    '融资租入固定资产',
    '现金的期末余额',
    '减：现金的期初余额',
    '加：现金等价物的期末余额',
    '减：现金等价物的期初余额',
    '现金及现金等价物净增加额',
    '流动比率(非金融类指标)',
    '速动比率(非金融类指标)',
    '现金比率(%)(非金融类指标)',
    '利息保障倍数(非金融类指标)',
    '非流动负债比率(%)(非金融类指标)',
    '流动负债比率(%)(非金融类指标)',
    '现金到期债务比率(%)(非金融类指标)',
    '有形资产净值债务率(%)',
    '权益乘数(%)',
    '股东的权益/负债合计(%)',
    '有形资产/负债合计(%)',
    '经营活动产生的现金流量净额/负债合计(%)(非金融类指标)',
    'EBITDA/负债合计(%)(非金融类指标)',
    '应收帐款周转率(非金融类指标)',
    '存货周转率(非金融类指标)',
    '运营资金周转率(非金融类指标)',
    '总资产周转率(非金融类指标)',
    '固定资产周转率(非金融类指标)',
    '应收帐款周转天数(非金融类指标)',
    '存货周转天数(非金融类指标)',
    '流动资产周转率(非金融类指标)',
    '流动资产周转天数(非金融类指标)',
    '总资产周转天数(非金融类指标)',
    '股东权益周转率(非金融类指标)',
    '营业收入增长率(%)',
    '净利润增长率(%)',
    '净资产增长率(%)',
    '固定资产增长率(%)',
    '总资产增长率(%)',
    '投资收益增长率(%)',
    '营业利润增长率(%)',
    '扣非每股收益同比(%)',
    '扣非净利润同比(%)',
    '暂无',
    '成本费用利润率(%)',
    '营业利润率(非金融类指标)',
    '营业税金率(非金融类指标)',
    '营业成本率(非金融类指标)',
    '净资产收益率',
    '投资收益率',
    '销售净利率(%)',
    '总资产报酬率',
    '净利润率(非金融类指标)',
    '销售毛利率(%)(非金融类指标)',
    '三费比重(非金融类指标)',
    '管理费用率(非金融类指标)',
    '财务费用率(非金融类指标)',
    '扣除非经常性损益后的净利润',
    '息税前利润(EBIT)',
    '息税折旧摊销前利润(EBITDA)',
    'EBITDA/营业总收入(%)(非金融类指标)',
    '资产负债率(%)',
    '流动资产比率(非金融类指标)',
    '货币资金比率(非金融类指标)',
    '存货比率(非金融类指标)',
    '固定资产比率',
    '负债结构比(非金融类指标)',
    '归属于母公司股东权益/全部投入资本(%)',
    '股东的权益/带息债务(%)',
    '有形资产/净债务(%)',
    '每股经营性现金流(元)',
    '营业收入现金含量(%)(非金融类指标)',
    '经营活动产生的现金流量净额/经营活动净收益(%)',
    '销售商品提供劳务收到的现金/营业收入(%)',
    '经营活动产生的现金流量净额/营业收入',
    '资本支出/折旧和摊销',
    '每股现金流量净额(元)',
    '经营净现金比率（短期债务）(非金融类指标)',
    '经营净现金比率（全部债务）',
    '经营活动现金净流量与净利润比率',
    '全部资产现金回收率',
    '营业收入',
    '营业利润',
    '归属于母公司所有者的净利润',
    '扣除非经常性损益后的净利润',
    '经营活动产生的现金流量净额',
    '投资活动产生的现金流量净额',
    '筹资活动产生的现金流量净额',
    '现金及现金等价物净增加额',
    '总股本',
    '已上市流通A股',
    '已上市流通B股',
    '已上市流通H股',
    '股东人数(户)',
    '第一大股东的持股数量',
    '十大流通股东持股数量合计(股)',
    '十大股东持股数量合计(股)',
    '机构总量（家）',
    '机构持股总量(股)',
    'QFII机构数',
    'QFII持股量',
    '券商机构数',
    '券商持股量',
    '保险机构数',
    '保险持股量',
    '基金机构数',
    '基金持股量',
    '社保机构数',
    '社保持股量',
    '私募机构数',
    '私募持股量',
    '财务公司机构数',
    '财务公司持股量',
    '年金机构数',
    '年金持股量',
    '十大流通股东中持有A股合计(股)',
    '第一大流通股东持股量(股)',
    '自由流通股(股)',
    '受限流通A股(股)',
    '一般风险准备(金融类)',
    '其他综合收益(利润表)',
    '综合收益总额(利润表)',
    '归属于母公司股东权益(资产负债表)',
    '银行机构数(家)(机构持股)',
    '银行持股量(股)(机构持股)',
    '一般法人机构数(家)(机构持股)',
    '一般法人持股量(股)(机构持股)',
    '近一年净利润(元)',
    '信托机构数(家)(机构持股)',
    '信托持股量(股)(机构持股)',
    '特殊法人机构数(家)(机构持股)',
    '特殊法人持股量(股)(机构持股)',
    '加权净资产收益率(每股指标)',
    '扣非每股收益(单季度财务指标)',
    '最近一年营业收入（万元）',
    '国家队持股数量（万股)',
    '业绩预告-本期净利润同比增幅下限%',
    '业绩预告-本期净利润同比增幅上限%',
    '归母净利润（业绩快报）',
    '扣非净利润（业绩快报）',
    '总资产（业绩快报）',
    '净资产（业绩快报）',
    '每股收益（业绩快报）',
    '摊薄净资产收益率（业绩快报）',
    '加权净资产收益率（业绩快报）',
    '每股净资产（业绩快报）',
    '应付票据及应付账款(资产负债表)',
    '应收票据及应收账款(资产负债表)',
    '递延收益(资产负债表)',
    '其他综合收益(资产负债表)',
    '其他权益工具(资产负债表)',
    '其他收益(利润表)',
    '资产处置收益(利润表)',
    '持续经营净利润(利润表)',
    '终止经营净利润(利润表)',
    '研发费用(利润表)',
    '其中:利息费用(利润表-财务费用)',
    '其中:利息收入(利润表-财务费用)',
    '近一年经营活动现金流净额',
    '近一年归母净利润（万元）',
    '近一年扣非净利润（万元）',
    '近一年现金净流量（万元）',
    '基本每股收益（单季度）',
    '营业总收入(单季度)(万元)',
    '业绩预告公告日期 ',
    '财报公告日期',
    '业绩快报公告日期',
    '近一年投资活动现金流净额(万元)',
    '业绩预告-本期净利润下限(万元)',
    '业绩预告-本期净利润上限(万元)',
    '营业总收入TTM(万元)',
    '员工总数(人)',
    '每股企业自由现金流',
    '每股股东自由现金流',
    'col323',
    'col324',
    'col325',
    'col326',
    'col327',
    'col328',
    'col329',
    'col330',
    'col331',
    'col332',
    'col333',
    'col334',
    'col335',
    'col336',
    'col337',
    'col338',
    'col339',
    'col340',
    'col341',
    'col342',
    'col343',
    'col344',
    'col345',
    'col346',
    'col347',
    'col348',
    'col349',
    'col350',
    'col351',
    'col352',
    'col353',
    'col354',
    'col355',
    'col356',
    'col357',
    'col358',
    'col359',
    'col360',
    'col361',
    'col362',
    'col363',
    'col364',
    'col365',
    'col366',
    'col367',
    'col368',
    'col369',
    'col370',
    'col371',
    'col372',
    'col373',
    'col374',
    'col375',
    'col376',
    'col377',
    'col378',
    'col379',
    'col380',
    'col381',
    'col382',
    'col383',
    'col384',
    'col385',
    'col386',
    'col387',
    'col388',
    'col389',
    'col390',
    'col391',
    'col392',
    'col393',
    'col394',
    'col395',
    'col396',
    'col397',
    'col398',
    'col399',
    'col400',
    '专项储备(万元)',
    '结算备付金(万元)',
    '拆出资金(万元)',
    '发放贷款及垫款(万元)(流动资产科目)',
    '衍生金融资产(万元)',
    '应收保费(万元)',
    '应收分保账款(万元)',
    '应收分保合同准备金(万元)',
    '买入返售金融资产(万元)',
    '划分为持有待售的资产(万元)',
    '发放贷款及垫款(万元)(非流动资产科目)',
    '向中央银行借款(万元)',
    '吸收存款及同业存放(万元)',
    '拆入资金(万元)',
    '衍生金融负债(万元)',
    '卖出回购金融资产款(万元)',
    '应付手续费及佣金(万元)',
    '应付分保账款(万元)',
    '保险合同准备金(万元)',
    '代理买卖证券款(万元)',
    '代理承销证券款(万元)',
    '划分为持有待售的负债(万元)',
    '预计负债(万元)',
    '递延收益(万元)（流动负债科目）',
    '其中:优先股(万元)(非流动负债科目)',
    '永续债(万元)(非流动负债科目)',
    '长期应付职工薪酬(万元)',
    '其中:优先股(万元)(所有者权益科目)',
    '永续债(万元)(所有者权益科目)',
    '债权投资(万元)',
    '其他债权投资(万元)',
    '其他权益工具投资(万元)',
    '其他非流动金融资产(万元)',
    '合同负债(万元)',
    '合同资产(万元)',
    '其他资产(万元)',
    '应收款项融资(万元)',
    '使用权资产(万元)',
    '租赁负债(万元)',
    'col440',
    'col441',
    'col442',
    'col443',
    'col444',
    'col445',
    'col446',
    'col447',
    'col448',
    'col449',
    'col450',
    'col451',
    'col452',
    'col453',
    'col454',
    'col455',
    'col456',
    'col457',
    'col458',
    'col459',
    'col460',
    'col461',
    'col462',
    'col463',
    'col464',
    'col465',
    'col466',
    'col467',
    'col468',
    'col469',
    'col470',
    'col471',
    'col472',
    'col473',
    'col474',
    'col475',
    'col476',
    'col477',
    'col478',
    'col479',
    'col480',
    'col481',
    'col482',
    'col483',
    'col484',
    'col485',
    'col486',
    'col487',
    'col488',
    'col489',
    'col490',
    'col491',
    'col492',
    'col493',
    'col494',
    'col495',
    'col496',
    'col497',
    'col498',
    'col499',
    'col500',
    '稀释每股收益(元)',
    '营业总收入(万元)',
    '汇兑收益(万元)',
    '其中:归属于母公司综合收益(万元)',
    '其中:归属于少数股东综合收益(万元)',
    '利息收入(万元)',
    '已赚保费(万元)',
    '手续费及佣金收入(万元)',
    '利息支出(万元)',
    '手续费及佣金支出(万元)',
    '退保金(万元)',
    '赔付支出净额(万元)',
    '提取保险合同准备金净额(万元)',
    '保单红利支出(万元)',
    '分保费用(万元)',
    '其中:非流动资产处置利得(万元)',
    '信用减值损失(万元)',
    '净敞口套期收益(万元)',
    '营业总成本(万元)',
    '信用减值损失(万元、2019格式)',
    '资产减值损失(万元、2019格式)',
    'col522',
    'col523',
    'col524',
    'col525',
    'col526',
    'col527',
    'col528',
    'col529',
    'col530',
    'col531',
    'col532',
    'col533',
    'col534',
    'col535',
    'col536',
    'col537',
    'col538',
    'col539',
    'col540',
    'col541',
    'col542',
    'col543',
    'col544',
    'col545',
    'col546',
    'col547',
    'col548',
    'col549',
    'col550',
    'col551',
    'col552',
    'col553',
    'col554',
    'col555',
    'col556',
    'col557',
    'col558',
    'col559',
    'col560',
    '加:其他原因对现金的影响2(万元)(现金的期末余额科目)',
    '客户存款和同业存放款项净增加额(万元)',
    '向中央银行借款净增加额(万元)',
    '向其他金融机构拆入资金净增加额(万元)',
    '收到原保险合同保费取得的现金(万元)',
    '收到再保险业务现金净额(万元)',
    '保户储金及投资款净增加额(万元)',
    '处置以公允价值计量且其变动计入当期损益的金融资产净增加额(万元)',
    '收取利息、手续费及佣金的现金(万元)',
    '拆入资金净增加额(万元)',
    '回购业务资金净增加额(万元)',
    '客户贷款及垫款净增加额(万元)',
    '存放中央银行和同业款项净增加额(万元)',
    '支付原保险合同赔付款项的现金(万元)',
    '支付利息、手续费及佣金的现金(万元)',
    '支付保单红利的现金(万元)',
    '其中:子公司吸收少数股东投资收到的现金(万元)',
    '其中:子公司支付给少数股东的股利、利润(万元)',
    '投资性房地产的折旧及摊销(万元)',
    '信用减值损失(万元)',
]



import time
from datetime import datetime
from pathlib import Path

from tdxpy.reader import CustomerBlockReader

from mootdx.consts import TYPE_FLATS
from mootdx.consts import TYPE_GROUP
from mootdx.logger import logger
from mootdx.utils import get_stock_market

try:
    from time import time_ns
except ImportError:
    # 兼容 3.6
    time_ns = lambda: int(time.time() * 1e9)  # noqa


class Customize:
    items: dict = {}

    def __init__(self, tdxdir=None):
        self.vipdoc = Path(tdxdir, 'T0002', 'blocknew')
        self.tdxdir = str(tdxdir)

    def create(self, name: str = None, symbol: list = None, **kwargs):
        return _blocknew(self.tdxdir, name=name, symbol=symbol, **kwargs)

    def remove(self, name: str = None):
        """
        删除板块数据

        :param name:
        :return:
        """

        block_data = self.search()

        if block_data.empty:
            logger.error('自定义板块数据是空的')
            return

        block_file = Path(self.vipdoc) / 'blocknew.cfg'
        block_temp = block_data[block_data.blockname == name]
        block_type = ''

        # 删除blk文件
        if block_temp.block_type.to_list():
            block_type = list(set(block_temp.block_type.to_list()))[0]
            [
                Path(self.vipdoc, f'{x}.blk').unlink()
                for x in block_temp.block_type.to_list()
                if Path(self.vipdoc, f'{x}.blk').is_file()
            ]

        # 读取文件
        block_data = Path(block_file).read_bytes().decode(encoding='gb2312')

        # 替换内容
        data = name + ((50 - len(name.encode('gbk', 'ignore'))) * '\x00')
        data += block_type + ((70 - len(block_type.encode('gbk', 'ignore'))) * '\x00')
        data = block_data.replace(data, '')
        data = bytes(data.encode('gbk', 'ignore'))

        # 写回文件
        return Path(block_file).write_bytes(data)

    def search(self, name: str = None, group=False):
        """
        按名称查找自定义板块名称

        :param name:
        :param group:
        :return:
        """

        if name:
            result = CustomerBlockReader().get_df(str(self.vipdoc), TYPE_GROUP)
            result = result[result.blockname == name]

            if result.empty:
                return None

            result = result.code_list.values
            result = list(set(result[0].split(',')))

            return result

        # 全部数据
        return CustomerBlockReader().get_df(str(self.vipdoc), (TYPE_FLATS, TYPE_GROUP)[group])

    def update(self, name: str = None, symbol=None, overflow=False):
        """
        修改自定义板块内容

        :param name: 板块名称
        :param symbol: 股票代码, ['600036','600016']
        :param overflow:
        """

        if not name:
            return False

        # 板块路径
        block_path = self.vipdoc
        block_code = list(symbol)

        # 板块数据
        block_data = self.search()
        block_temp = block_data[block_data.blockname == name]

        # 对于名称空的情况, 直接创建写入
        if block_temp.empty:
            logger.debug(f'block_temp is empty {block_temp.empty}')
            return _blocknew(self.tdxdir, name=name, symbol=list(set(symbol)))

        # 覆盖情况
        if not overflow:
            block_code += block_temp.code.to_list()

        # 取 blk 文件名, block_type 不为空
        if block_temp.block_type.to_list():
            block_type = list(set(block_temp.block_type.to_list()))[0]
            logger.debug(f'发现板块文件: {block_type}')
        else:
            # block_type 为空的话
            block_type = datetime.now().strftime('%Y%m%d%H%M%S')
            logger.debug(f'板块文件找不到: {block_type}')

        # 去重股票代码
        block_code = list(set(block_code))
        logger.debug(f'证券代码: {block_code}')

        # 股票代码逗号隔开拼字符串
        block_code = '\n'.join([f'{get_stock_market(s)}{s}' for s in block_code])

        # 写入 blk 文件
        block_file = Path(block_path, f'{block_type}.blk')
        logger.debug(f'写入文件 : {block_file}')

        return block_file.write_text(block_code, encoding='gb2312')


def _blocknew(tdxdir: str = None, name: str = None, symbol: list = None, blk_file: str = None, **kwargs):  # noqa
    """
    自定义模块写入函数

    :param tdxdir: tdx 路径
    :param name: 自定义板块名称
    :param symbol: 自定义板块股票代码集合
    :param blk_file: 自定义板块股票代码集合文件
    :return: bool
    """

    if not tdxdir:
        logger.error(f'通达信路径不存在或者空: {tdxdir}')
        return False

    # 自定义板块名称未传入则自动按时间生成名称
    if not name:
        name = datetime.now().strftime('%Y%m%d%H%M%S')

    # 按时间生成 blk 文件名
    blk_file = blk_file if blk_file else str(time_ns())

    vipdoc = Path(tdxdir, 'T0002', 'blocknew')
    symbol = list(set(symbol))

    # 判断目录是否存在
    if not Path(vipdoc).is_dir():
        logger.error(f'自定义板块目录错误: {vipdoc}')
        return False

    block_file = Path(vipdoc) / 'blocknew.cfg'

    # 文件不存在就创建
    if not block_file.exists():
        block_file.write_text('')

    # 判断名字是否重名
    with open(block_file, 'rb') as fp:
        names = fp.read().decode('gbk', 'ignore')
        names = names.split('\x00')

        names = [x for x in names if x != '']
        names = [v for i, v in enumerate(names) if i % 2 == 0]

        if name in names:
            # todo symbol 不空则合并, 空则删除
            logger.error('自定义板块名称重复.')
            raise Exception('自定义板块名称重复.')

    # 写 blk 文件
    with open(f'{vipdoc}/{blk_file}.blk', 'w') as fp:
        fp.write('\n'.join([f'{get_stock_market(s)}{s}' for s in symbol]))

    # 写 blocknew.cfg 文件
    with open(block_file, 'ab') as fp:
        data = name + ((50 - len(name.encode('gbk', 'ignore'))) * '\x00')
        data += blk_file + ((70 - len(blk_file.encode('gbk', 'ignore'))) * '\x00')

        data = bytes(data.encode('gbk', 'ignore'))
        fp.write(data)

    return True



from pathlib import Path

import pandas as pd
from tdxpy.reader import BlockReader

from mootdx.consts import TYPE_FLATS
from mootdx.consts import TYPE_GROUP
from mootdx.logger import logger


class BaseParse:
    def __init__(self, tdxdir):  # noqa
        self.tdxdir = tdxdir  # noqa

    def parse(self, symbol=None, group=False, **kwargs):  # noqa
        """
        获取板块数据

        参考: http://blog.sina.com.cn/s/blog_623d2d280102vt8y.html

        :param symbol:  板块文件
        :param group:   分组解析
        :return: pd.dataFrame or None
        """

        suffix = Path(symbol).suffix or '.dat'
        symbol = Path(symbol).stem

        vipdoc = (Path('T0002', 'hq_cache'), '')['incon' in symbol]  # noqa
        vipdoc = Path(vipdoc, f'{symbol}{suffix}')  # noqa

        if not Path(self.tdxdir, vipdoc).exists():
            logger.error(f'文件不存在: {vipdoc}')
            return None

        if 'incon' in symbol:  # noqa
            return self.__incon(vipdoc)

        if 'block_' in symbol and suffix == '.dat':
            return BlockReader().get_df(str(Path(self.tdxdir, vipdoc)), (TYPE_FLATS, TYPE_GROUP)[bool(group)])

        return self.cfg(vipdoc)

    def read_text(self, path):
        return Path(self.tdxdir, path).read_text(encoding='gbk').strip()

    def __incon(self, path):  # noqa
        t = self.read_text(path)
        m = [x for x in t.split('######')]
        v = [n.split() for n in m if n.strip()]

        d = {i[0]: [c.split('|') for c in i[1:]] for i in v}
        d = {key: dict([vv for vv in val if len(vv) == 2]) for key, val in d.items()}

        return d

    def cfg(self, path):
        ts = self.read_text(path)
        ls = [ll.split('|') for ll in ts.split()]

        return pd.DataFrame(ls)



import logging
import os
import warnings
from pathlib import Path

from mootdx import __version__
from mootdx.logger import logger
from mootdx.utils import get_config_path
from mootdx.utils import to_file

try:
    import click
    from prettytable import PrettyTable
except (ImportError, ModuleNotFoundError):
    logging.basicConfig(level=logging.WARNING)
    warnings.warn('!!! 缺少命令行依赖, 请使用次命令进行安装: pip install "mootdx[cli]"', DeprecationWarning)
    logging.warning('!!! 缺少命令行依赖, 请使用次命令进行安装: pip install "mootdx[cli]"')
    exit(-1)


@click.group()
@click.version_option(__version__, '-V', '--version', prog_name='Mootdx', message='%(prog)s: v%(version)s')
@click.help_option('-h', '--help')
def entry():
    ...


@entry.command(help='读取股票在线行情数据.')
@click.help_option('-h', '--help')
@click.option('-o', '--output', default=None, help='输出文件, 支持CSV, HDF5, Excel等格式.')
@click.option('-s', '--symbol', default='600000', help='股票代码.')
@click.option('-a', '--action', default='bars', help='操作类型 (daily: 日线, minute: 一分钟线, fzline: 五分钟线).', )
@click.option('-m', '--market', default='std', help='证券市场, 默认 std (std: 标准股票市场, ext: 扩展市场).')
def quotes(symbol, action, market, output):
    from mootdx.quotes import Quotes

    client = Quotes.factory(market=market, multithread=True)

    try:
        action = 'bars' if 'daily' else action
        if action == 'daily':
            frequency = 9
        elif action == 'minute':
            frequency = 8
        elif action == 'fzline':
            frequency = 0
        else:
            frequency = 9

        feed = getattr(client, 'bars')(symbol=symbol, frequency=frequency)
        to_file(feed, output) if output else None
        click.echo(feed)
    except Exception as e:
        raise e


@entry.command(help='读取股票本地行情数据.')
@click.help_option('-h', '--help')
@click.option('-d', '--tdxdir', default='C:/new_tdx', help='通达信数据目录.')
@click.option('-s', '--symbol', default='600000', help='股票代码.')
@click.option('-a', '--action', default='daily', help='操作类型 (daily: 日线, minute: 一分钟线, fzline: 五分钟线).')
@click.option('-m', '--market', default='std', help='证券市场, 默认 std (std: 标准股票市场, ext: 扩展市场).')
@click.option('-o', '--output', default=None, help='输出文件, 支持 CSV, HDF5, Excel 等格式.')
def reader(symbol, action, market, tdxdir, output):
    from mootdx.reader import Reader

    client = Reader.factory(market=market, tdxdir=tdxdir)

    try:
        feed = getattr(client, action)(symbol=symbol)
        to_file(feed, output) if output else None
        click.echo(feed)
    except Exception as e:
        raise e


@entry.command(help='测试行情服务器.', name='bestip')
@click.help_option('-h', '--help')
@click.option('-l', '--limit', default=5, help='显示最快前几个，默认 5.')
@click.option('-v', '--verbose', count=True, help='详细模式')
def server(limit, verbose):
    from mootdx.server import bestip

    if verbose:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        logger.addHandler(ch)
        logger.setLevel(logging.DEBUG)

    bestip(limit=limit, console=True, sync=False)

    config = get_config_path('config.json')
    logger.info(f'[√] 已经将最优服务器IP写入配置文件 {config}')


@entry.command(help='财务文件下载&解析.')
@click.help_option('-h', '--help')
@click.option('-p', '--parse', default=None, help='要解析文件名')
@click.option('-f', '--fetch', default=None, help='下载财务文件的文件名')
@click.option('-a', '--downall', is_flag=True, help='下载全部文件')
@click.option('-o', '--output', default=None, help='输出文件, 支持 CSV, HDF5, Excel, JSON 等格式.')
@click.option('-d', '--downdir', default='output', help='下载文件目录')
@click.option('-l', '--listfile', is_flag=True, default=False, help='显示全部文件')
@click.option('-v', '--verbose', count=True, help='详细模式')
def affair(parse, fetch, downdir, output, downall, verbose, listfile):
    from mootdx.affair import Affair

    if verbose:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        logger.addHandler(ch)
        logger.setLevel(logging.DEBUG)

    files = Affair.files()

    if listfile:
        t = PrettyTable(['filename', 'filesize', 'hash'])
        t.align['filename'] = 'l'
        t.align['filesize'] = 'l'
        t.align['hash'] = 'l'
        t.padding_width = 0

        for x in files:
            t.add_row([x['filename'], x['filesize'], x['hash']])

        click.echo(t)
        return

    # download all files
    if downall or fetch == 'all':
        feed = Affair.fetch(downdir=downdir)
        return to_file(feed, output) if output else None

    # download file
    if fetch:
        return Affair.fetch(downdir=downdir, filename=fetch.strip('.zip') + '.zip')

    # parse file
    if parse:
        parse = parse.strip('.zip') + '.zip'
        files = [x['filename'] for x in files]

        if parse in files:
            # 文件不存在，先下载
            Path(downdir, parse).exists() or Affair.fetch(downdir=downdir, filename=parse)
            feed = Affair.parse(downdir=downdir, filename=parse)
            output and to_file(feed, output)
            click.echo(feed)
        else:
            logger.error('没找到要解析的文件.')


@entry.command(help='批量下载行情数据.')
@click.help_option('-h', '--help')
@click.option('-o', '--output', default='bundle', help='转存文件目录.')
@click.option('-s', '--symbol', default='600000', help='股票代码. 多个用,隔开')
@click.option('-a', '--action', default='bars', help='操作类型 (daily: 日线, minute: 一分钟线, fzline: 五分钟线).')
@click.option('-m', '--market', default='std', help='证券市场, 默认 std (std: 标准股票市场, ext: 扩展市场).')
@click.option('-e', '--extension', default='csv', help='转存文件的格式, 支持 CSV, HDF5, Excel, JSON 等格式.')
def bundle(symbol, action, market, output, extension):
    """
    批量下载行情数据
    :return:
    """
    from mootdx.quotes import Quotes

    client = Quotes.factory(market=market, multithread=True)
    symbol = symbol.replace('，', ',').strip(',').split(',')

    for code in symbol:
        try:
            if action == 'daily':
                frequency = 9
            elif action == 'minute':
                frequency = 8
            elif action == 'fzline':
                frequency = 0
            else:
                frequency = 9

            feed = getattr(client, 'bars')(symbol=code, frequency=frequency)
            output and to_file(feed, os.path.join(output, f'{code}.{extension}'))
            click.echo('下载完成 {}'.format(code))
        except Exception as e:
            raise e

    click.echo(f'[√] 下载文件到 "{os.path.realpath(output)}"')


if __name__ == '__main__':
    entry()



import pandas as pd


def _reversion(bfq_data, xdxr_data, type_):
    """使用数据库数据进行复权"""
    info = xdxr_data.query('category==1')
    bfq_data = bfq_data.assign(if_trade=1)

    if len(info) > 0:
        # 有除权数据
        data = pd.concat([bfq_data, info.loc[bfq_data.index[0]: bfq_data.index[-1], ['category']]], axis=1)
        data['if_trade'].fillna(value=0, inplace=True)

        data = data.fillna(method='ffill')
        data = pd.concat(
            [data, info.loc[bfq_data.index[0]: bfq_data.index[-1], ['fenhong', 'peigu', 'peigujia', 'songzhuangu']]],
            axis=1)
    else:
        data = pd.concat([bfq_data, info.loc[:, ['category', 'fenhong', 'peigu', 'peigujia', 'songzhuangu']]], axis=1)

    # 数据补全
    data = data.fillna(0)

    # 计算前日收盘
    data['preclose'] = (data['close'].shift(1) * 10 - data['fenhong'] + data['peigu'] * data['peigujia']) / (
        10 + data['peigu'] + data['songzhuangu'])

    # 前复权
    if type_.lower() in ['01', 'qfq']:
        data['adj'] = (data['preclose'].shift(-1) / data['close']).fillna(1)[::-1].cumprod()
        # ohlc 数据进行复权计算
        for col in ['open', 'high', 'low', 'close', 'preclose']:
            data[col] = data[col] * data['adj']

    # 后复权
    if type_.lower() in ['02', 'hfq']:
        data['adj'] = (data['preclose'].shift(-1) / data['close']).fillna(1).cumprod()
        for col in ['open', 'high', 'low', 'close', 'preclose']:
            data[col] = data[col] / data['adj']

    # data["volume"] = data.get("volume", data.get("vol"))
    data['volume'] = data['volume'] / data['adj']
    # data['volume'] = data['volume'] / data['adj'] if 'volume' in data.columns else data['vol'] / data['adj']

    try:
        # 大该是涨跌幅
        data['high_limit'] = data['high_limit'] * data['adj']
        data['low_limit'] = data['low_limit'] * data['adj']
    except:
        pass

    data = data.query('if_trade==1 and open != 0')
    data = data.drop(['fenhong', 'peigu', 'peigujia', 'songzhuangu', 'if_trade', 'category'], axis=1, errors='ignore')

    return data


def reversion(stock_data, xdxr, type_='01'):
    def _fetch_xdxr(collections=None):
        """获取股票除权信息数据"""
        columns = [
            'category',
            'category_meaning',
            'date',
            'fenhong',
            'fenshu',
            'liquidity_after',
            'liquidity_before',
            'name',
            'peigu',
            'peigujia',
            'shares_after',
            'shares_before',
            'songzhuangu',
            'suogu',
            'xingquanjia',
        ]

        try:
            data = collections[['category', 'fenhong', 'peigu', 'peigujia', 'songzhuangu']]
            data['date'] = pd.to_datetime(data[['year', 'month', 'day']], utc=False)
            data = data.set_index(['date'])

            # data = data.drop(['year', 'month', 'day', ], axis=1)
            # data = pd.DataFrame([item for item in collections.find({"code": symbol})]).drop(["_id"], axis=1)
            # data = collections
            # data["date"] = pd.to_datetime(data["date"], utc=False)
            # data["date"] = pd.to_datetime(xdxr[["year", "month", "day"]], utc=False)
            # return data.set_index(["date", "code"], drop=False)
            return data
        except:
            return pd.DataFrame(data=[], columns=columns)

    # '股票 日线/分钟线 动态复权接口'
    # if isinstance(stock_data.index, pd.MultiIndex):
    #     symbol = stock_data.index.remove_unused_levels().levels[1][0]
    # else:
    #     symbol = stock_data["code"][0]
    # symbol = ''
    # symbol = (
    #     stock_data.index.remove_unused_levels().levels[1][0]
    #     if isinstance(stock_data.index, pd.MultiIndex)
    #     else stock_data["code"][0]
    # )
    return _reversion(bfq_data=stock_data, xdxr_data=_fetch_xdxr(xdxr), type_=type_)


# 算法一样
def baoli_qfq(df, xdxr):
    peigu = xdxr['peigu']  # 配股
    fenhong = xdxr['fenhong']  # 分红
    peigujia = xdxr['peigujia']  # 配股价
    songzhuangu = xdxr['songzhuangu']  # 送转股

    for i in range(0, len(xdxr)):
        fh = fenhong[i]
        pg = peigu[i]
        pgj = peigujia[i]
        szg = songzhuangu[i]
        date = xdxr.index[i]

        df.loc[df.index < date, 'close'] = (df['close'][df.index < date] * 10 - fh + pg * pgj) / (10 + pg + szg)
        df.loc[df.index < date, 'open'] = (df['open'][df.index < date] * 10 - fh + pg * pgj) / (10 + pg + szg)
        df.loc[df.index < date, 'high'] = (df['high'][df.index < date] * 10 - fh + pg * pgj) / (10 + pg + szg)
        df.loc[df.index < date, 'low'] = (df['low'][df.index < date] * 10 - fh + pg * pgj) / (10 + pg + szg)

    return df



import time
from functools import wraps


def timeit(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        delta = (time.time() - start_time)

        if delta > 1:
            t = ' '.join([str(delta), 's'])
        else:
            t = ' '.join([str(delta * 1000), 'ms'])

        print(' -> function', func.__name__, 'time:', t)
        return result

    return decorator



import socket
import threading

from tdxpy.base_socket_client import BaseSocketClient
from tdxpy.base_socket_client import CONNECT_TIMEOUT
from tdxpy.base_socket_client import TrafficStatSocket
from tdxpy.exceptions import TdxConnectionError
from tdxpy.heartbeat import HeartBeatThread
from tdxpy.reader import TdxDailyBarReader

from mootdx.logger import logger


class MooTdxDailyBarReader(TdxDailyBarReader):
    """感谢 bopomofo 的鼎力支持"""

    SECURITY_TYPE = [
        'SH_A_STOCK',
        'SH_B_STOCK',
        'SH_STAR_STOCK',
        'SH_INDEX',
        'SH_FUND',
        'SH_BOND',
        'SZ_A_STOCK',
        'SZ_B_STOCK',
        'SZ_INDEX',
        'SZ_FUND',
        'SZ_BOND',
    ]

    SECURITY_COEFFICIENT = {
        'SH_A_STOCK': [0.01, 0.01],
        'SH_B_STOCK': [0.001, 0.01],
        'SH_STAR_STOCK': [0.01, 0.01],
        'SH_INDEX': [0.01, 1.0],
        'SH_FUND': [0.001, 1.0],
        'SH_BOND': [0.001, 1.0],
        'SZ_A_STOCK': [0.01, 0.01],
        'SZ_B_STOCK': [0.01, 0.01],
        'SZ_INDEX': [0.01, 1.0],
        'SZ_FUND': [0.001, 0.01],
        'SZ_BOND': [0.001, 0.01],
    }

    def get_security_type(self, fname):

        exchange = str(fname[-12:-10]).lower()
        code_head = fname[-10:-8]

        if exchange == self.SECURITY_EXCHANGE[0]:
            if code_head in ['00', '30']:
                return 'SZ_A_STOCK'

            if code_head in ['20']:
                return 'SZ_B_STOCK'

            if code_head in ['39']:
                return 'SZ_INDEX'

            if code_head in ['15', '16']:
                return 'SZ_FUND'

            if code_head in ['10', '11', '12', '13', '14']:
                return 'SZ_BOND'
        elif exchange == self.SECURITY_EXCHANGE[1]:
            if code_head in ['60']:
                return 'SH_A_STOCK'

            if code_head in ['90']:
                return 'SH_B_STOCK'

            if code_head in ['68']:
                return 'SH_STAR_STOCK'

            if code_head in ['00', '88', '99']:
                return 'SH_INDEX'

            if code_head in ['50', '51', '58']:
                return 'SH_FUND'

            if code_head in ['01', '10', '11', '12', '13', '14']:
                return 'SH_BOND'
        else:
            logger.error('Unknown security exchange !\n')
            raise NotImplementedError


class MooBaseSocketClient(BaseSocketClient):
    def __init__(self):
        super().__init__()
        self.client = None

    def connect(self, ip='101.227.73.20', port=7709, time_out=CONNECT_TIMEOUT, bindport=None, bindip='0.0.0.0'):
        """

        :param ip:  服务器ip 地址
        :param port:  服务器端口
        :param time_out: 连接超时时间
        :param bindport: 绑定的本地端口
        :param bindip: 绑定的本地ip
        :return: 是否连接成功 True/False
        """

        self.client = TrafficStatSocket(socket.AF_INET, socket.SOCK_STREAM)
        self.client.settimeout(time_out)

        logger.debug('connecting to server : %s on port :%d' % (ip, port))

        try:
            self.ip = ip
            self.port = port

            if bindport is not None:
                self.client.bind((bindip, bindport))

            self.client.connect((ip, port))
        except socket.timeout as e:
            logger.debug(e)
            logger.debug('connection expired')

            if self.raise_exception:
                raise TdxConnectionError('connection timeout error')

            return False
        except Exception as e:
            logger.debug(e)
            if self.raise_exception:
                raise TdxConnectionError('other errors')

            return False

        logger.debug('connected!')

        if self.need_setup:
            self.setup()

        if self.heartbeat:
            self.stop_event = threading.Event()
            self.heartbeat_thread = HeartBeatThread(self, self.stop_event, self.heartbeat_interval)
            self.heartbeat_thread.start()

        return self

    def disconnect(self):

        if self.heartbeat_thread and self.heartbeat_thread.is_alive():
            self.stop_event.set()

        if self.client:
            logger.debug('disconnecting')

            try:
                self.client.shutdown(socket.SHUT_RDWR)
                self.client.close()
                self.client = None
            except Exception as e:
                logger.debug(str(e))

                if self.raise_exception:
                    raise TdxConnectionError('disconnect err')

            logger.debug('disconnected')

    def setup(self):
        pass
